/*********************************
 * This file is part of the "Interactive Fan Chart Demo"
 * by Geoffrey M. Draper.
 * 
 * Geoffrey M. Draper, author of this work, hereby grants
 * everyone permission to use and distribute this code,
 * with or without modifications for any purpose,
 * commercial or otherwise. 
 * 
 * It is requested, although not required, that derivative
 * works include an acknowledgment somewhere in the
 * derivative work's credits.
 *********************************/

package fanchart;


import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.geom.Point2D;

public class CurvedText {

	private RadialCache cache;
	private Dimension windowSize;
	private float radius0, radius1;
	private Color color;

	public CurvedText(RadialCache rc, Dimension windowSize) {
		this(rc, windowSize, Config.textColor);
	}

	public CurvedText(RadialCache rc, Dimension windowSize, Color color) {
		cache = rc;
		this.windowSize = windowSize;
		this.color = color;
	}

	private void setup(int i, Graphics2D g) {
		g.setColor(color);
		g.setFont(Config.regularFont);
		radius0 = windowSize.height/2 - cache.getVertex(i-1, 0).y;
		radius1 = windowSize.height/2 - cache.getVertex(i, 0).y;
	}

	//this method is as ugly as all getout.
	//Make it understandable otherwise it will be impossible to maintain.
	//TODO consider converting it to draw 3 lines of text?
	public void draw(final String text, int i, float availableTicks,
			float middleIndex, Graphics2D g) {
		//draw anywhere on the screen
		//plus negotiate font size based on available space
		boolean upsideDown = false;
		float topRadius = 0;
		float bottomRadius = 0;
		int labelLengthInPixels;
		float index;
		float numberOfTicksForLabel;
		String topLine;
		String bottomLine = "";
		boolean topGood = false;
		boolean twoLineMode = false;
		Font topFont = Config.regularFont, bottomFont = Config.regularFont;
		setup(i, g);
		int textHeight = g.getFontMetrics(topFont).getHeight();
		topRadius = radius1-(radius1-radius0+textHeight/2)/2;
		float circumference = (float)(2*topRadius*Math.PI);
		float proportionOfCircumferenceForLabel;
		topLine = new String(text);
		labelLengthInPixels = g.getFontMetrics(topFont).stringWidth(topLine);
		proportionOfCircumferenceForLabel = labelLengthInPixels/circumference;
		numberOfTicksForLabel = (RadialCache.RESOLUTION * proportionOfCircumferenceForLabel);
		if (numberOfTicksForLabel > availableTicks) {
			twoLineMode = true;
			topGood = false;
		} else {
			topGood = true;
			twoLineMode = false;
		}
		while (!topGood) {
			String currentText = topLine;
			String attempt = car(currentText);
			if (attempt != null) {
				topLine = attempt.trim();
				bottomLine = (cdr(currentText) + " " + bottomLine).trim();
				labelLengthInPixels = g.getFontMetrics(topFont).stringWidth(topLine);
				proportionOfCircumferenceForLabel = labelLengthInPixels/circumference;
				numberOfTicksForLabel = (RadialCache.RESOLUTION * proportionOfCircumferenceForLabel);
				if (numberOfTicksForLabel <= availableTicks) {
					topGood = true;
				}
			} else {
				topGood = true;
			}
		}
		if (cache.getVertex(0, (int)middleIndex).y > windowSize.height/2) {
			upsideDown = true;
		}
		if (twoLineMode) {
			bottomRadius = topRadius - textHeight/2;
			topRadius += textHeight/2;
		}
		String textToRender;
		g.setFont(topFont);
		int charactersToUse = topLine.length();
		do {
			textToRender = topLine.substring(0, charactersToUse);
			labelLengthInPixels = g.getFontMetrics(topFont).stringWidth(textToRender);
			proportionOfCircumferenceForLabel = labelLengthInPixels/circumference;
			numberOfTicksForLabel = (RadialCache.RESOLUTION * proportionOfCircumferenceForLabel);
			--charactersToUse;
		} while (numberOfTicksForLabel > availableTicks && charactersToUse >= 0);
		index = middleIndex - numberOfTicksForLabel/2f;
		drawString(textToRender, i, index, topRadius, labelLengthInPixels, numberOfTicksForLabel, upsideDown, g);
		if (twoLineMode) {
			g.setFont(bottomFont);
			charactersToUse = bottomLine.length();
			do {
				textToRender = bottomLine.substring(0, charactersToUse);
				labelLengthInPixels = g.getFontMetrics(bottomFont).stringWidth(textToRender);
				proportionOfCircumferenceForLabel = labelLengthInPixels/circumference;
				numberOfTicksForLabel = (RadialCache.RESOLUTION * proportionOfCircumferenceForLabel);
				--charactersToUse;
			} while (numberOfTicksForLabel > availableTicks && charactersToUse >= 0);
			index = middleIndex - numberOfTicksForLabel/2f;
			drawString(textToRender, i, index, bottomRadius, labelLengthInPixels, numberOfTicksForLabel, upsideDown, g);
		}
	}

	private void drawString(String text, int i, float index, float radius,
			float labelLengthInPixels, float numberOfTicksForLabel,
			boolean upsideDown, Graphics2D g) {

		//I define all these variables outside the loop
		//for efficiency's sake.
		String currentChar = "";
		int letterLength;
		float numberOfTicksForCurrentChar;
		Point2D.Float p1, p2;
		Point2D.Float tangent = new Point2D.Float();
		double angle;
		Point drawHere = new Point();

		int fudge = 1;	//do we add or subtract ticks from the index?
		double fudge2 = 0;	//do we flip the angle or not?
		float ratio = (radius-radius0) / (radius1-radius0);

		//by using the fudge values, we avoid having to perform if's
		//inside the for loop, below.
		if (upsideDown) {
			index += numberOfTicksForLabel;
			fudge = -1;
			ratio = (radius1-radius) / (radius1-radius0);
			fudge2 = Math.PI;
		}


		for (int j=0; j<text.length(); ++j) {
			currentChar = Character.toString(text.charAt(j));
			letterLength = g.getFontMetrics().stringWidth(currentChar);
			numberOfTicksForCurrentChar = ((float)letterLength/labelLengthInPixels) * numberOfTicksForLabel;
			p1 = cache.getVertex(i-1, (int)index);
			p2 = cache.getVertex(i, (int)index);
			index += (numberOfTicksForCurrentChar*fudge);
			tangent.setLocation(p2.x-p1.x, p2.y-p1.y);
			angle = Math.PI - Math.atan2(tangent.x, tangent.y) + fudge2;
			drawHere.setLocation((int)(p1.x + ratio*tangent.x), (int)(p1.y + ratio*tangent.y));
			g.rotate(angle, drawHere.x, drawHere.y);
			g.drawString(currentChar, drawHere.x,
					drawHere.y);				
			g.rotate(-angle, drawHere.x, drawHere.y);
		}

	}

	private String car(String s) {
		int i = s.lastIndexOf(' ');
		if (i == -1) {
			return null;
		}
		return s.substring(0, i);
	}

	private String cdr(String s) {
		int i = s.lastIndexOf(' ');
		return s.substring(i+1);
	}
	
}
