/*********************************
 * This file is part of the "Interactive Fan Chart Demo"
 * by Geoffrey M. Draper.
 * 
 * Geoffrey M. Draper, author of this work, hereby grants
 * everyone permission to use and distribute this code,
 * with or without modifications for any purpose,
 * commercial or otherwise. 
 * 
 * It is requested, although not required, that derivative
 * works include an acknowledgment somewhere in the
 * derivative work's credits.
 *********************************/

package fanchart;




import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics2D;
import java.awt.Point;
import java.util.Map;

public class PersonWedge {
	private PersonWedge motherWedge, fatherWedge, childWedge;
	private Person entity;
	protected HyrumPolygon mainPolygon;
	private ForceAncestorsWedge forceAncestorsWedge;
	protected float start;
	protected float length;
	protected float idealStart;
	protected float idealLength;
	protected float textAnchorIndex;
	protected int radius1, radius2;
	private int generation;
	private boolean isRoot;
	private boolean collapsed;
	private boolean expanded;
	private boolean forceAncestors;
	//protected static final Point2D.Float windowCenter = new Point2D.Float();//TODO delete?
	protected static final float MIN_THICKNESS = RadialCache.RESOLUTION/360f;
	private static final long serialVersionUID = 42L;
	private Color color;
	private static int highestVisibleGeneration = Integer.MIN_VALUE;
	
	public PersonWedge(Person f, float start, float length, int rad) {
		this(f, start, length, rad, null, null);
	}

	public PersonWedge(Person f, float start, float length, int rad, PersonWedge fw, PersonWedge mw) {
		this.entity = f;
		fatherWedge = fw;
		motherWedge = mw;
		collapsed = false;
		expanded = false;
		forceAncestors = false;
		forceAncestorsWedge = null;
		childWedge = null;
		refresh(start, start, length, rad);
	}
	
	public void refresh(float start, float alternateStart, float length, int rad) {
		if (expanded) {
			this.start = alternateStart;
			this.length = length*2;
		} else {
			this.start = start;
			this.length = length;
		}
		idealLength = length;
		idealStart = start;
		mainPolygon = new HyrumPolygon();
		if (rad == 0) {//TODO fix magic number!
			isRoot = true;
		} else {
			isRoot = false;
		}
		radius1 = rad;
		radius2 = rad+1;
		generation = radius2;
		if (generation > highestVisibleGeneration) {
			highestVisibleGeneration = generation;
		}
		color = Config.getColorForPerson(entity);
		
		createAncestralWedges();
	}

	public void createAncestralWedges() {
		if (generation < Config.numberOfGenerationsToShow || forceAncestors) {
			createAncestralWedges_helper();
		} else {
			fatherWedge = null; motherWedge = null;
			if (entity.getFather() != null || entity.getMother() != null) {
				forceAncestorsWedge = new ForceAncestorsWedge(start, length);
			}
		}
	}
	
	private void createAncestralWedges_helper() {
		if (entity.getFather() == null) {
			fatherWedge = null;
		} else {
			if (fatherWedge == null) {
				fatherWedge = new PersonWedge(entity.getFather(), start, length/2f, radius2);
			} else {
				fatherWedge.refresh(start, start, length/2f, radius2);
			}
			fatherWedge.childWedge = this;
		}
		if (entity.getMother() == null) {
			motherWedge = null;
		} else {
			if (motherWedge == null) {
				motherWedge = new PersonWedge(entity.getMother(), start+length/2f, length/2f, radius2);
			} else {
				motherWedge.refresh(start+length/2f, start, length/2f, radius2);
			}
			motherWedge.childWedge = this;
		}
	}
	
	public void buildVertices(RadialCache cache, Dimension windowSize, int offset) {
		int res = RadialCache.RESOLUTION;
		mainPolygon.reset();

		int index0 = Math.round(offset + start);
		if (index0 < 0) index0 += res;
		int llength = Math.round(length);
		textAnchorIndex = offset + start + length/2;

		if (!collapsed) {
			if (!isRoot) {
				for (int i=0; i<=llength; ++i) {	
					mainPolygon.addPoint(cache.getVertex(radius1, (index0 + i) % res));
				}
			}
			for (int i=llength; i>=0; --i) {
				mainPolygon.addPoint(cache.getVertex(radius2, (index0 + i) % res));
			}
			mainPolygon.closePath();
		}

		if (!collapsed) {
			if (fatherWedge != null) fatherWedge.buildVertices(cache, windowSize, offset);
			if (motherWedge != null) motherWedge.buildVertices(cache, windowSize, offset);
			if (forceAncestorsWedge != null) forceAncestorsWedge.buildVertices(cache, radius2, offset);
		}
	}

	public void draw(CurvedText textRenderer, Graphics2D g, boolean drawText) {
		if (!collapsed) {
			boolean mayDrawTextThisTime = drawText;
			if (isRoot) {
				mayDrawTextThisTime = !FanChartWidget.currentlyAnimatingNewRoot;
			}
			g.setColor(color);
			mainPolygon.fill(g);			

			//draw dark outline around wedge
			g.setColor(Config.wedgeBorderColor);
			g.setStroke(Config.regularLine);
			mainPolygon.draw(g);
			if (mayDrawTextThisTime) {
				textRenderer.draw(entity.getName(), radius2, length, textAnchorIndex, g);
			}

			if (fatherWedge != null) fatherWedge.draw(textRenderer, g, drawText);
			if (motherWedge != null) motherWedge.draw(textRenderer, g, drawText);
			if (forceAncestorsWedge != null) forceAncestorsWedge.draw(g);
		}
	}

	public void drawThickOutline(Graphics2D g) {
		g.setColor(Config.wedgeBorderColor);
		g.setStroke(Config.thickLine);
		mainPolygon.draw(g);
	}

	public void drawThickOutlineForExpandedWedges(Graphics2D g) {
		if (expanded) {
			drawThickOutline(g);
		}	
		if (fatherWedge != null) fatherWedge.drawThickOutlineForExpandedWedges(g);
		if (motherWedge != null) motherWedge.drawThickOutlineForExpandedWedges(g);
	}

	public boolean containsPoint(Point p) {
		//inefficient hackery
		boolean result = isPointInsideForceGenerationWedge(p);
		if (!result) {
			result = mainPolygon.contains(p);
		}
		return result;
	}

	public PersonWedge findClickedWedge(Point p) {
		PersonWedge result = null;
		if (containsPoint(p)) {
			result = this;
		} else {
			if (fatherWedge != null) {
				result = fatherWedge.findClickedWedge(p);
			}
			if (result == null && motherWedge != null) {
				result = motherWedge.findClickedWedge(p);
			}
		}
		return result;
	}

	public Person getEntity() {
		return entity;
	}

	public Person getPerson() {
		return getEntity();
	}

	@Override
	public String toString() {
		return entity.getName();
	}

	public String getName() {
		return entity.getName();
	}

	float getLength() {
		return length;
	}

	float getStart() {
		return start;
	}

	float getIdealStart() {
		return idealStart;
	}

	protected void reset() {
		length = idealLength;
		start = idealStart;
		createAncestralWedges();
	}


	public void resetStart(float s) {
		start = s;
	}

	public void adjustLengthByDelta(float delta) {
		length += delta;
	}

	public void resetLength(float len) {
		this.length = len;
	}

	public String getTooltipText(Point pos) {
		if (containsPoint(pos))
			return entity.getName();
		else
			return null;
	}

	public String getTooltipText() {
		return entity.getName();
	}

	public boolean isRoot() {
		return isRoot;
	}

	public void setAsRoot(boolean isRoot) {
		this.isRoot = isRoot;
	}

	public void populateList(Map<Integer, PersonWedge> wedges) {
		wedges.put(entity.getID(), this);
		if (fatherWedge != null) fatherWedge.populateList(wedges);
		if (motherWedge != null) motherWedge.populateList(wedges);
	}

	public boolean isCollapsed() {
		return collapsed;
	}

	public void setCollapsed(boolean collapsed) {
		this.collapsed = collapsed;
		if (collapsed) {
			motherWedge = null;
			fatherWedge = null;
			forceAncestors = false;
		} else {
			createAncestralWedges();
		}
	}

	public PersonWedge getSpouseWedge() {
		PersonWedge result = null;
		if (childWedge.motherWedge.equals(this)) {
			result = childWedge.fatherWedge;
		} else {
			result = childWedge.motherWedge;
		}
		return result;
	}

	public PersonWedge getChildWedge() {
		return childWedge;
	}

	public boolean isExpanded() {
		return expanded;
	}

	public void setNotExpanded() {
		expanded = false;
	}

	public void setExpanded() {
		setExpanded(getSpouseWedge().getIdealStart());
	}

	public void setExpanded(float otherStart) {
		expanded = true;
		if (expanded) {
			length = 2f*idealLength;
			if (otherStart < start) {
				start = otherStart;
			}
			createAncestralWedges();
		}
	}

	@Override
	public boolean equals(Object other) {
		boolean result = false;
		if (other instanceof PersonWedge) {
			result = entity.equals(((PersonWedge)other).entity);
		}
		return result;
	}

	public HyrumPolygon copyPolygon() {
		return new HyrumPolygon(mainPolygon);
	}
	
	public boolean isPointInsideForceGenerationWedge(Point p) {
		boolean result = false;
		if (forceAncestorsWedge != null) {
			result = forceAncestorsWedge.containsPoint(p);
		}
		return result;
	}
	
	public void forceNextGeneration() {
		if (Config.numberOfGenerationsToShow <= generation) {
			++Config.db_totalNumberOfRadii;
		}
		forceAncestorsWedge = null;
		perpetuateForceAncestors(true);
		createAncestralWedges_helper();
	}
	
	private void perpetuateForceAncestors(boolean flag) {
		forceAncestors = flag;
		if (childWedge != null) {
			childWedge.perpetuateForceAncestors(flag);
		}
	}
	
	public static int getHighestVisibleGeneration() {
		return highestVisibleGeneration;
	}
	


}
